/*
 * scisort - sort lines by numbers in scientific notation
 *
 * LICENCE:
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License
 *  as published by the Free Software Foundation; either version 2
 *  of the License, or (at your option) any later version.
 *  See http://www.gnu.org/copyleft/gpl.html
 *
 * AUTHOR:
 *  Written by Freek van Walderveen, july 2005
 *
 */


#include <iostream>
#include <algorithm>
#include <vector>
#include <cstdlib>
#include <cstdio>
#include <cstring>
#include <unistd.h>

//#define ISSPACE(c) std::isspace((c))
#define ISSPACE(c) ((c)==' '||(c)=='\t')

int key;
int reverse;

class Line {
  public:
    Line(char *s) {
      unsigned int i = 0;
      unsigned int len = strlen(s);
      m_val = 0.0;
      
      m_s = s;
      while (i < len && ISSPACE(s[i])) {
        i++;
      }
      if (i >= len) {
        // empty line, sort as first
        m_empty = 1;
      } else {
        m_empty = 0;
      }
      for (int k = 0; k < key; k++) {
        while (i < len && !ISSPACE(s[i])) {
          i++;
        }
        while (i < len && ISSPACE(s[i])) {
          i++;
        }
        if (i >= len) {
          // semi-empty line
          m_empty = 1;
        }
      }
      if (i < len) {
        m_val = strtod(&(s[i]), NULL);
      }
    }
    
    inline double value() const { return m_val; }
    inline char *line() const { return m_s; }
    inline int empty() const { return m_empty; }
    
    inline int
    operator<(const Line& right) const {
      return reverse ^ (!right.empty() && (value() < right.value() || empty()));
    }
    
  private:
    char *m_s;
    double m_val;
    int m_empty;
};

void usage (void) {
  std::cout << "scisort (20050711) by Freek van Walderveen (http://vanwal.nl/scisort/)" << std::endl;
  std::cout << "Usage: scisort [-k <key>] [-s] [-r]" << std::endl << std::endl;
  std::cout << "scisort reads from stdin and writes to stdout." << std::endl;
  std::cout << "Use -r for reverse sort and -k <key> for sorting by column <key> (first column is 1). Relative order of elements that compare equal is preserved when using -s (stable sort)." << std::endl;
  exit(0);
}

int main (int argc, char *argv[]) {
  std::vector<Line> numbers;
  char *curline = NULL;
  size_t n;
  int stable = 0;
  // for option parsing
  int optchr;
  extern char *optarg;
  
  // parse options
  key = 1;
  reverse = 0;
  while ((optchr = getopt(argc, argv, "h?srk:")) != -1) {
  	switch (optchr) {
	    case 'r':
        reverse++;
        break;
	    case 's':
        stable++;
        break;
	    case 'k':
        key = atol(optarg) - 1;
        if (key < 1) key = 0;
        break;
      default: // -h, -? and all the rest
        usage();
    }
  }
  
  // read stdin
  //std::cerr << "Reading..." << std::endl;
  while (getline(&curline, &n, stdin) >= 0) {
    numbers.push_back(Line(curline));
    curline = NULL;
  }
  
  // sort
  //std::cerr << "Sorting..." << std::endl;
  if (stable) {
    std::stable_sort(numbers.begin(), numbers.end());
  } else {
    std::sort(numbers.begin(), numbers.end());
  }
  
  // print sorted
  //std::cerr << "Printing..." << std::endl;
  for (std::vector<Line>::iterator it = numbers.begin(); it < numbers.end(); ++it) {
    std::cout << it->line();
    free(it->line());
  }
}
